body {
    font-family: Arial, sans-serif;
    margin: 20px;
}

header {
    background-color: #f8f9fa;
    padding: 10px;
    border-bottom: 1px solid #ccc;
}

nav {
    display: flex;
    justify-content: flex-start;
}

nav a {
    text-decoration: none;
    color: #007bff;
    padding: 10px 15px;
    margin-right: 10px;
    border-radius: 5px;
    transition: background-color 0.3s;
}

nav a:hover {
    background-color: #e2e6ea;
}

nav a.active {
    background-color: #007bff;
    color: white;
}

#main {
    display: flex;
    position: relative;
    justify-content: center;
}

#output {
    max-width: 900px;
    padding: 20px;
    margin-top: 20px;
    box-shadow: 2px 0 5px rgba(0, 0, 0, 0.1);
    background-color: #f4f4f4;
    flex-grow: 1;
}

blockquote {
    border-left: 8px solid #fafafa;
    padding: 1rem;
    border: 1px solid #999;
    page-break-inside: avoid;
}

pre {
    background-color: #fafafa;
    border-radius: 6px;
    padding: 8px;
    line-height: 1.25;
    overflow-x: auto;
}

h1,
h2,
h3 {
    border-bottom: 2px solid #fafafa;
    margin-bottom: 1.15rem;
    padding-bottom: 0.5rem;
    text-align: center;
}

figure {
    margin: 1.5em 0;
    text-align: center;
}

figure img {
    max-width: 100%;
    height: auto;
    border: 1px solid #ddd;
    border-radius: 4px;
}
